package cn.sh.ideal.util;

import org.apache.commons.codec.binary.Base64;

import javax.crypto.Cipher;
import java.io.ByteArrayOutputStream;
import java.nio.charset.StandardCharsets;
import java.security.*;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.X509EncodedKeySpec;
import java.util.HashMap;
import java.util.Map;

/**
 * Base64报错 需要引入以下依赖
 * <dependency>
 * <groupId>commons-codec</groupId>
 * <artifactId>commons-codec</artifactId>
 * <version>1.10</version>
 * </dependency>
 */
public class RSAUtil {

    /**
     * RSA最大加密明文大小
     */
    private static final int MAX_ENCRYPT_BLOCK = 117;

    /**
     * RSA最大解密密文大小
     */
    private static final int MAX_DECRYPT_BLOCK = 128;

    // 获取公钥的key
    private static final String PUBLIC_KEY = "RSAPublicKey";

    // 获取私钥的key
    private static final String PRIVATE_KEY = "RSAPrivateKey";

    /**
     * 获取密钥对
     *
     * @return 密钥对
     */
    public static Map<String, Object> getKeyPair() throws Exception {
        KeyPairGenerator generator = KeyPairGenerator.getInstance("RSA");
        generator.initialize(1024);
        KeyPair keyPair = generator.generateKeyPair();
        // 保存公私钥
        Map<String, Object> map = new HashMap<>();
        map.put(PUBLIC_KEY, keyPair.getPublic());
        map.put(PRIVATE_KEY, keyPair.getPrivate());
        return map;
    }

    /**
     * 在秘钥对中取出公钥
     *
     * @param keyPair 秘钥对
     * @return 公钥字符串
     */
    public static String getPublic(Map<String, Object> keyPair) {
        PublicKey publicKey = (PublicKey) keyPair.get(PUBLIC_KEY);
        return new String(Base64.encodeBase64(publicKey.getEncoded()));
    }

    /**
     * 在秘钥对中取出私钥
     *
     * @param keyPair 秘钥对
     * @return 私钥字符串
     */
    public static String getPrivate(Map<String, Object> keyPair) {
        PrivateKey privateKey = (PrivateKey) keyPair.get(PRIVATE_KEY);
        return new String(Base64.encodeBase64(privateKey.getEncoded()));
    }

    /**
     * 通过私钥字符串 获取私钥
     *
     * @param privateKey 私钥字符串
     * @return PrivateKey
     */
    public static PrivateKey getPrivateKey(String privateKey) throws Exception {
        KeyFactory keyFactory = KeyFactory.getInstance("RSA");
        byte[] decodedKey = Base64.decodeBase64(privateKey.getBytes());
        PKCS8EncodedKeySpec keySpec = new PKCS8EncodedKeySpec(decodedKey);
        return keyFactory.generatePrivate(keySpec);
    }

    /**
     * 通过公钥字符串 获取公钥
     *
     * @param publicKey 公钥字符串
     * @return PublicKey
     */
    public static PublicKey getPublicKey(String publicKey) throws Exception {
        KeyFactory keyFactory = KeyFactory.getInstance("RSA");
        byte[] decodedKey = Base64.decodeBase64(publicKey.getBytes());
        X509EncodedKeySpec keySpec = new X509EncodedKeySpec(decodedKey);
        return keyFactory.generatePublic(keySpec);
    }

    /**
     * RSA加密
     *
     * @param data      待加密数据
     * @param publicKey 公钥
     * @return 密文
     */
    public static String encrypt(String data, PublicKey publicKey) throws Exception {
        Cipher cipher = Cipher.getInstance("RSA/ECB/PKCS1Padding");
        cipher.init(Cipher.ENCRYPT_MODE, publicKey);
        int inputLen = data.getBytes().length;
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        int offset = 0;
        byte[] cache;
        int i = 0;
        // 对数据分段加密
        while (inputLen - offset > 0) {
            if (inputLen - offset > MAX_ENCRYPT_BLOCK) {
                cache = cipher.doFinal(data.getBytes(), offset, MAX_ENCRYPT_BLOCK);
            } else {
                cache = cipher.doFinal(data.getBytes(), offset, inputLen - offset);
            }
            out.write(cache, 0, cache.length);
            i++;
            offset = i * MAX_ENCRYPT_BLOCK;
        }
        byte[] encryptedData = out.toByteArray();
        out.close();
        // 获取加密内容使用base64进行编码,并以UTF-8为标准转化成字符串
        // 加密后的字符串
        return Base64.encodeBase64String(encryptedData);
    }

    /**
     * RSA解密
     *
     * @param data       待解密数据
     * @param privateKey 私钥
     * @return
     */
    public static String decrypt(String data, PrivateKey privateKey) throws Exception {
        Cipher cipher = Cipher.getInstance("RSA/ECB/PKCS1Padding");
        cipher.init(Cipher.DECRYPT_MODE, privateKey);
        byte[] dataBytes = Base64.decodeBase64(data);
        int inputLen = dataBytes.length;
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        int offset = 0;
        byte[] cache;
        int i = 0;
        // 对数据分段解密
        while (inputLen - offset > 0) {
            if (inputLen - offset > MAX_DECRYPT_BLOCK) {
                cache = cipher.doFinal(dataBytes, offset, MAX_DECRYPT_BLOCK);
            } else {
                cache = cipher.doFinal(dataBytes, offset, inputLen - offset);
            }
            out.write(cache, 0, cache.length);
            i++;
            offset = i * MAX_DECRYPT_BLOCK;
        }
        out.close();
        // 解密后的内容
        return out.toString(String.valueOf(StandardCharsets.UTF_8));
    }

    /**
     * 签名
     *
     * @param data       待签名数据
     * @param privateKey 私钥
     * @return 签名
     */
    public static String sign(String data, PrivateKey privateKey) throws Exception {
        byte[] keyBytes = privateKey.getEncoded();
        PKCS8EncodedKeySpec pkcs8KeySpec = new PKCS8EncodedKeySpec(keyBytes);
        KeyFactory keyFactory = KeyFactory.getInstance("RSA");
        PrivateKey key = keyFactory.generatePrivate(pkcs8KeySpec);
        Signature signature = Signature.getInstance("MD5withRSA");
        signature.initSign(key);
        signature.update(data.getBytes());
        return new String(Base64.encodeBase64(signature.sign()));
    }

    /**
     * 验签
     *
     * @param srcData   原始字符串
     * @param publicKey 公钥
     * @param sign      签名
     * @return 是否验签通过
     */
    public static boolean verify(String srcData, PublicKey publicKey, String sign) throws Exception {
        byte[] keyBytes = publicKey.getEncoded();
        X509EncodedKeySpec keySpec = new X509EncodedKeySpec(keyBytes);
        KeyFactory keyFactory = KeyFactory.getInstance("RSA");
        PublicKey key = keyFactory.generatePublic(keySpec);
        Signature signature = Signature.getInstance("MD5withRSA");
        signature.initVerify(key);
        signature.update(srcData.getBytes());
        return signature.verify(Base64.decodeBase64(sign.getBytes()));
    }

    public static void main(String[] args) throws Exception {
//        // 生成密钥对
        Map<String, Object> keyPair = RSAUtil.getKeyPair();
        String aPublic = RSAUtil.getPublic(keyPair);
        String aPrivate = RSAUtil.getPrivate(keyPair);
        System.out.println("公钥:" + aPublic);
        System.out.println("私钥:" + aPrivate);

        String data = "17636665565";
        String encrypt = RSAUtil.encrypt(data, RSAUtil.getPublicKey(aPublic));
        System.out.println("加密后: " + encrypt);

        String decrypt = RSAUtil.decrypt(encrypt, RSAUtil.getPrivateKey(aPrivate));
        System.out.println("解密后: " + decrypt);

        String sign = getMobilePhone("15", "4545151515", "ajsdnjkasndjaksndad", "asndjkad", aPrivate);
        System.out.println("生成sign: " + sign);
    }

    /**
     * sdk取号接口
     * 规则: userId + appId + 一键登录token + appSecret
     * @param appId appId
     * @param appSecret appSecret
     * @param userId 用户id
     * @param token 一键登录token
     * @param aPrivate RSA私钥
     * @return sign
     */
    public static String getMobilePhone(String userId, String appId, String token, String appSecret, String aPrivate) {
        try {
            String str = userId + appId + token + appSecret;
            return RSAUtil.sign(str, RSAUtil.getPrivateKey(aPrivate));
        } catch (Exception ignored) {
        }

        return "";
    }

}